@@ -33,7 +33,7 @@
 #include "tiled.h"
 #include "codecs/decoder.h"
 #include "color-conversion/colorconversion.h"
-#include "api/libheif/api_structs.h"
+#include "api_structs.h"
 #include "plugin_registry.h"
 #include "security_limits.h"
 
@@ -61,6 +61,17 @@ ImageItem::ImageItem(HeifContext* context, heif_item_id id)
 }
 
 
+bool ImageItem::is_property_essential(const std::shared_ptr<Box>& property) const
+{
+  if (property->get_short_type() == fourcc("ispe")) {
+    return is_ispe_essential();
+  }
+  else {
+    return property->is_essential();
+  }
+}
+
+
 std::shared_ptr<HeifFile> ImageItem::get_file() const
 {
   return m_heif_context->get_heif_file();
@@ -69,6 +80,10 @@ std::shared_ptr<HeifFile> ImageItem::get_file() const
 
 heif_property_id ImageItem::add_property(std::shared_ptr<Box> property, bool essential)
 {
+  if (!property) {
+    return 0;
+  }
+
   // TODO: is this correct? What happens when add_property does deduplicate the property?
   m_properties.push_back(property);
   return get_file()->add_property(get_id(), property, essential);
@@ -77,20 +92,15 @@ heif_property_id ImageItem::add_property(std::shared_ptr<Box> property, bool ess
 
 heif_property_id ImageItem::add_property_without_deduplication(std::shared_ptr<Box> property, bool essential)
 {
+  if (!property) {
+    return 0;
+  }
+
   m_properties.push_back(property);
   return get_file()->add_property_without_deduplication(get_id(), property, essential);
 }
 
 
-Error ImageItem::init_decoder_from_item(heif_item_id id)
-{
-  m_id = id;
-
-  Error err = on_load_file();
-  return err;
-}
-
-
 heif_compression_format ImageItem::compression_format_from_fourcc_infe_type(uint32_t type)
 {
   switch (type) {
@@ -172,6 +182,12 @@ std::shared_ptr<ImageItem> ImageItem::alloc_for_infe_box(HeifContext* ctx, const
   else if (item_type == fourcc("j2k1")) {
     return std::make_shared<ImageItem_JPEG2000>(ctx, id);
   }
+  else if (item_type == fourcc("lhv1")) {
+    return std::make_shared<ImageItem_Error>(item_type, id,
+                                             Error{heif_error_Unsupported_feature,
+                                                   heif_suberror_Unsupported_image_type,
+                                                   "Layered HEVC images (lhv1) are not supported yet"});
+  }
   else if (item_type == fourcc("mski")) {
     return std::make_shared<ImageItem_mask>(ctx, id);
   }
@@ -207,6 +223,8 @@ std::shared_ptr<ImageItem> ImageItem::alloc_for_compression_format(HeifContext*
       return std::make_shared<ImageItem_AVIF>(ctx);
     case heif_compression_VVC:
       return std::make_shared<ImageItem_VVC>(ctx);
+    case heif_compression_AVC:
+      return std::make_shared<ImageItem_AVC>(ctx);
 #if WITH_UNCOMPRESSED_CODEC
     case heif_compression_uncompressed:
       return std::make_shared<ImageItem_uncompressed>(ctx);
@@ -223,25 +241,28 @@ std::shared_ptr<ImageItem> ImageItem::alloc_for_compression_format(HeifContext*
 }
 
 
-Result<ImageItem::CodedImageData> ImageItem::encode_to_bitstream_and_boxes(const std::shared_ptr<HeifPixelImage>& image,
-                                                                           struct heif_encoder* encoder,
-                                                                           const struct heif_encoding_options& options,
-                                                                           enum heif_image_input_class input_class)
+Result<Encoder::CodedImageData> ImageItem::encode_to_bitstream_and_boxes(const std::shared_ptr<HeifPixelImage>& image,
+                                                                           heif_encoder* encoder,
+                                                                           const heif_encoding_options& options,
+                                                                           heif_image_input_class input_class)
 {
   // === generate compressed image bitstream
 
-  Result<ImageItem::CodedImageData> encodeResult = encode(image, encoder, options, input_class);
-  if (encodeResult.error) {
+  Result<Encoder::CodedImageData> encodeResult = encode(image, encoder, options, input_class);
+  if (!encodeResult) {
     return encodeResult;
   }
 
-  CodedImageData& codedImage = encodeResult.value;
+  Encoder::CodedImageData& codedImage = *encodeResult;
 
   // === generate properties
 
   // --- choose which color profile to put into 'colr' box
 
-  add_color_profile(image, options, input_class, options.output_nclx_profile, codedImage);
+  auto colr_boxes = add_color_profile(image, options, input_class, options.output_nclx_profile);
+  codedImage.properties.insert(codedImage.properties.end(),
+                               colr_boxes.begin(),
+                               colr_boxes.end());
 
 
   // --- ispe
@@ -328,35 +349,15 @@ Result<ImageItem::CodedImageData> ImageItem::encode_to_bitstream_and_boxes(const
   }
   codedImage.properties.push_back(pixi);
 
+  // --- generate properties for image extra data
 
-  // --- write PASP property
+  // copy over ImageExtraData into image item
+  *static_cast<ImageExtraData*>(this) = static_cast<ImageExtraData>(*image);
 
-  if (image->has_nonsquare_pixel_ratio()) {
-    auto pasp = std::make_shared<Box_pasp>();
-    image->get_pixel_ratio(&pasp->hSpacing, &pasp->vSpacing);
-
-    codedImage.properties.push_back(pasp);
-  }
-
-
-  // --- write CLLI property
-
-  if (image->has_clli()) {
-    auto clli = std::make_shared<Box_clli>();
-    clli->clli = image->get_clli();
-
-    codedImage.properties.push_back(clli);
-  }
-
-
-  // --- write MDCV property
-
-  if (image->has_mdcv()) {
-    auto mdcv = std::make_shared<Box_mdcv>();
-    mdcv->mdcv = image->get_mdcv();
-
-    codedImage.properties.push_back(mdcv);
-  }
+  auto extra_data_properties = image->generate_property_boxes();
+  codedImage.properties.insert(codedImage.properties.end(),
+                               extra_data_properties.begin(),
+                               extra_data_properties.end());
 
   return encodeResult;
 }
@@ -364,9 +365,9 @@ Result<ImageItem::CodedImageData> ImageItem::encode_to_bitstream_and_boxes(const
 
 Error ImageItem::encode_to_item(HeifContext* ctx,
                                 const std::shared_ptr<HeifPixelImage>& image,
-                                struct heif_encoder* encoder,
-                                const struct heif_encoding_options& options,
-                                enum heif_image_input_class input_class)
+                                heif_encoder* encoder,
+                                const heif_encoding_options& options,
+                                heif_image_input_class input_class)
 {
   uint32_t input_width = image->get_width();
   uint32_t input_height = image->get_height();
@@ -376,12 +377,12 @@ Error ImageItem::encode_to_item(HeifContext* ctx,
 
   // compress image and assign data to item
 
-  Result<CodedImageData> codingResult = encode_to_bitstream_and_boxes(image, encoder, options, input_class);
-  if (codingResult.error) {
-    return codingResult.error;
+  Result<Encoder::CodedImageData> codingResult = encode_to_bitstream_and_boxes(image, encoder, options, input_class);
+  if (!codingResult) {
+    return codingResult.error();
   }
 
-  CodedImageData& codedImage = codingResult.value;
+  Encoder::CodedImageData& codedImage = *codingResult;
 
   auto infe_box = ctx->get_heif_file()->add_new_infe_box(get_infe_type());
   heif_item_id image_id = infe_box->get_item_ID();
@@ -392,9 +393,12 @@ Error ImageItem::encode_to_item(HeifContext* ctx,
 
   // set item properties
 
-  for (auto& propertyBox : codingResult.value.properties) {
+  for (auto& propertyBox : codingResult->properties) {
+    bool essential = is_property_essential(propertyBox);
+
+    // TODO: can we simply use add_property() ?
     int index = ctx->get_heif_file()->get_ipco_box()->find_or_append_child_box(propertyBox);
-    ctx->get_heif_file()->get_ipma_box()->add_property_for_item_ID(image_id, Box_ipma::PropertyAssociation{propertyBox->is_essential(),
+    ctx->get_heif_file()->get_ipma_box()->add_property_for_item_ID(image_id, Box_ipma::PropertyAssociation{essential,
                                                                                                            uint16_t(index + 1)});
   }
 
@@ -468,7 +472,7 @@ Error ImageItem::postprocess_coded_image_colorspace(heif_colorspace* inout_color
 
   if (*inout_colorspace == heif_colorspace_YCbCr) {
     auto nclx = get_color_profile_nclx();
-    if (nclx && nclx->get_matrix_coefficients() == 0) {
+    if (nclx.get_matrix_coefficients() == 0) {
       *inout_colorspace = heif_colorspace_RGB;
       *inout_chroma = heif_chroma_444; // TODO: this or keep the original chroma?
     }
@@ -480,8 +484,12 @@ Error ImageItem::postprocess_coded_image_colorspace(heif_colorspace* inout_color
 
 Error ImageItem::get_coded_image_colorspace(heif_colorspace* out_colorspace, heif_chroma* out_chroma) const
 {
-  auto decoder = get_decoder();
-  assert(decoder);
+  auto decoderResult = get_decoder();
+  if (!decoderResult) {
+    return decoderResult.error();
+  }
+
+  auto decoder = *decoderResult;
 
   Error err = decoder->get_coded_image_colorspace(out_colorspace, out_chroma);
   if (err) {
@@ -496,8 +504,12 @@ Error ImageItem::get_coded_image_colorspace(heif_colorspace* out_colorspace, hei
 
 int ImageItem::get_luma_bits_per_pixel() const
 {
-  auto decoder = get_decoder();
-  assert(decoder);
+  auto decoderResult = get_decoder();
+  if (!decoderResult) {
+    return decoderResult.error();
+  }
+
+  auto decoder = *decoderResult;
 
   return decoder->get_luma_bits_per_pixel();
 }
@@ -505,141 +517,41 @@ int ImageItem::get_luma_bits_per_pixel() const
 
 int ImageItem::get_chroma_bits_per_pixel() const
 {
-  auto decoder = get_decoder();
-  assert(decoder);
+  auto decoderResult = get_decoder();
+  if (!decoderResult) {
+    return decoderResult.error();
+  }
+
+  auto decoder = *decoderResult;
 
   return decoder->get_chroma_bits_per_pixel();
 }
 
 
-static std::shared_ptr<color_profile_nclx> compute_target_nclx_profile(const std::shared_ptr<HeifPixelImage>& image, const heif_color_profile_nclx* output_nclx_profile)
+Result<Encoder::CodedImageData> ImageItem::encode(const std::shared_ptr<HeifPixelImage>& image,
+                                                  heif_encoder* h_encoder,
+                                                  const heif_encoding_options& options,
+                                                  heif_image_input_class input_class)
 {
-  auto target_nclx_profile = std::make_shared<color_profile_nclx>();
-
-  // If there is an output NCLX specified, use that.
-  if (output_nclx_profile) {
-    target_nclx_profile->set_from_heif_color_profile_nclx(output_nclx_profile);
-  }
-    // Otherwise, if there is an input NCLX, keep that.
-  else if (auto input_nclx = image->get_color_profile_nclx()) {
-    *target_nclx_profile = *input_nclx;
-  }
-    // Otherwise, just use the defaults (set below)
-  else {
-    target_nclx_profile->set_undefined();
-  }
-
-  target_nclx_profile->replace_undefined_values_with_sRGB_defaults();
-
-  return target_nclx_profile;
+  auto encoder = get_encoder();
+  return encoder->encode(image, h_encoder, options, input_class);
 }
 
 
-static bool nclx_profile_matches_spec(heif_colorspace colorspace,
-                                      std::shared_ptr<const color_profile_nclx> image_nclx,
-                                      const struct heif_color_profile_nclx* spec_nclx)
+std::vector<std::shared_ptr<Box_colr> >
+ImageItem::add_color_profile(const std::shared_ptr<HeifPixelImage>& image,
+                             const heif_encoding_options& options,
+                             heif_image_input_class input_class,
+                             const heif_color_profile_nclx* target_heif_nclx)
 {
-  if (colorspace != heif_colorspace_YCbCr) {
-    return true;
-  }
+  std::vector<std::shared_ptr<Box_colr> > colr_boxes;
 
-  // No target specification -> always matches
-  if (!spec_nclx) {
-    return true;
-  }
-
-  if (!image_nclx) {
-    // if no input nclx is specified, compare against default one
-    image_nclx = std::make_shared<color_profile_nclx>();
-  }
-
-  if (image_nclx->get_full_range_flag() != (spec_nclx->full_range_flag == 0 ? false : true)) {
-    return false;
-  }
-
-  if (image_nclx->get_matrix_coefficients() != spec_nclx->matrix_coefficients) {
-    return false;
-  }
-
-  // TODO: are the colour primaries relevant for matrix-coefficients != 12,13 ?
-  //       If not, we should skip this test for anything else than matrix-coefficients != 12,13.
-  if (image_nclx->get_colour_primaries() != spec_nclx->color_primaries) {
-    return false;
-  }
-
-  return true;
-}
-
-
-Result<std::shared_ptr<HeifPixelImage>> ImageItem::convert_colorspace_for_encoding(const std::shared_ptr<HeifPixelImage>& image,
-                                                                                   struct heif_encoder* encoder,
-                                                                                   const struct heif_encoding_options& options)
-{
-  const heif_color_profile_nclx* output_nclx_profile;
-
-  if (const auto* nclx = get_forced_output_nclx()) {
-    output_nclx_profile = nclx;
-  }
-  else {
-    output_nclx_profile = options.output_nclx_profile;
-  }
-
-
-  heif_colorspace colorspace = image->get_colorspace();
-  heif_chroma chroma = image->get_chroma_format();
-
-  if (encoder->plugin->plugin_api_version >= 2) {
-    encoder->plugin->query_input_colorspace2(encoder->encoder, &colorspace, &chroma);
-  }
-  else {
-    encoder->plugin->query_input_colorspace(&colorspace, &chroma);
-  }
-
-
-  // If output format forces an NCLX, use that. Otherwise use user selected NCLX.
-
-  std::shared_ptr<color_profile_nclx> target_nclx_profile = compute_target_nclx_profile(image, output_nclx_profile);
-
-  // --- convert colorspace
-
-  std::shared_ptr<HeifPixelImage> output_image;
-
-  if (colorspace == image->get_colorspace() &&
-      chroma == image->get_chroma_format() &&
-      nclx_profile_matches_spec(colorspace, image->get_color_profile_nclx(), output_nclx_profile)) {
-    return image;
-  }
-
-
-  // @TODO: use color profile when converting
-  int output_bpp = 0; // same as input
-
-  //auto target_nclx = std::make_shared<color_profile_nclx>();
-  //target_nclx->set_from_heif_color_profile_nclx(target_heif_nclx);
-
-  auto output_image_result = convert_colorspace(image, colorspace, chroma, target_nclx_profile,
-                                                output_bpp, options.color_conversion_options,
-                                                get_context()->get_security_limits());
-  if (output_image_result.error) {
-    return output_image_result.error;
-  }
-
-  return *output_image_result;
-}
-
-
-void ImageItem::add_color_profile(const std::shared_ptr<HeifPixelImage>& image,
-                                  const struct heif_encoding_options& options,
-                                  enum heif_image_input_class input_class,
-                                  const heif_color_profile_nclx* target_heif_nclx,
-                                  ImageItem::CodedImageData& inout_codedImage)
-{
   if (input_class == heif_image_input_class_normal || input_class == heif_image_input_class_thumbnail) {
     auto icc_profile = image->get_color_profile_icc();
     if (icc_profile) {
       auto colr = std::make_shared<Box_colr>();
       colr->set_color_profile(icc_profile);
-      inout_codedImage.properties.push_back(colr);
+      colr_boxes.push_back(colr);
     }
 
 
@@ -664,44 +576,25 @@ void ImageItem::add_color_profile(const std::shared_ptr<HeifPixelImage>& image,
 
       auto colr = std::make_shared<Box_colr>();
       colr->set_color_profile(target_nclx_profile);
-      inout_codedImage.properties.push_back(colr);
+      colr_boxes.push_back(colr);
     }
   }
-}
 
-
-void ImageItem::CodedImageData::append(const uint8_t* data, size_t size)
-{
-  bitstream.insert(bitstream.end(), data, data + size);
-}
-
-
-void ImageItem::CodedImageData::append_with_4bytes_size(const uint8_t* data, size_t size)
-{
-  assert(size <= 0xFFFFFFFF);
-
-  uint8_t size_field[4];
-  size_field[0] = (uint8_t) ((size >> 24) & 0xFF);
-  size_field[1] = (uint8_t) ((size >> 16) & 0xFF);
-  size_field[2] = (uint8_t) ((size >> 8) & 0xFF);
-  size_field[3] = (uint8_t) ((size >> 0) & 0xFF);
-
-  bitstream.insert(bitstream.end(), size_field, size_field + 4);
-  bitstream.insert(bitstream.end(), data, data + size);
+  return colr_boxes;
 }
 
 
 Error ImageItem::transform_requested_tile_position_to_original_tile_position(uint32_t& tile_x, uint32_t& tile_y) const
 {
   Result<std::vector<std::shared_ptr<Box>>> propertiesResult = get_properties();
-  if (propertiesResult.error) {
-    return propertiesResult.error;
+  if (!propertiesResult) {
+    return propertiesResult.error();
   }
 
   heif_image_tiling tiling = get_heif_image_tiling();
 
   //for (auto& prop : std::ranges::reverse_view(propertiesResult.value)) {
-  for (auto propIter = propertiesResult.value.rbegin(); propIter != propertiesResult.value.rend(); propIter++) {
+  for (auto propIter = propertiesResult->rbegin(); propIter != propertiesResult->rend(); propIter++) {
     if (auto irot = std::dynamic_pointer_cast<Box_irot>(*propIter)) {
       switch (irot->get_rotation_ccw()) {
         case 90: {
@@ -750,8 +643,44 @@ Error ImageItem::transform_requested_tile_position_to_original_tile_position(uin
 }
 
 
-Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct heif_decoding_options& options,
-                                                                bool decode_tile_only, uint32_t tile_x0, uint32_t tile_y0) const
+void ImageItem::set_clli(const heif_content_light_level& clli)
+{
+  ImageExtraData::set_clli(clli);
+  add_property(get_clli_box(), false);
+}
+
+
+void ImageItem::set_mdcv(const heif_mastering_display_colour_volume& mdcv)
+{
+  ImageExtraData::set_mdcv(mdcv);
+  add_property(get_mdcv_box(), false);
+}
+
+
+void ImageItem::set_pixel_ratio(uint32_t h, uint32_t v)
+{
+  ImageExtraData::set_pixel_ratio(h, v);
+  add_property(get_pasp_box(), false);
+}
+
+
+void ImageItem::set_color_profile_nclx(const nclx_profile& profile)
+{
+  ImageExtraData::set_color_profile_nclx(profile);
+  add_property(get_colr_box_nclx(), false);
+}
+
+
+void ImageItem::set_color_profile_icc(const std::shared_ptr<const color_profile_raw>& profile)
+{
+  ImageExtraData::set_color_profile_icc(profile);
+  add_property(get_colr_box_icc(), false);
+}
+
+
+Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const heif_decoding_options& options,
+                                                                bool decode_tile_only, uint32_t tile_x0, uint32_t tile_y0,
+                                                                std::set<heif_item_id> processed_ids) const
 {
   // --- check whether image size (according to 'ispe') exceeds maximum
 
@@ -776,12 +705,16 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
 
   // --- decode image
 
-  Result<std::shared_ptr<HeifPixelImage>> decodingResult = decode_compressed_image(options, decode_tile_only, tile_x0, tile_y0);
-  if (decodingResult.error) {
-    return decodingResult.error;
+  Result<std::shared_ptr<HeifPixelImage>> decodingResult = decode_compressed_image(options, decode_tile_only, tile_x0, tile_y0, processed_ids);
+  if (!decodingResult) {
+    return decodingResult.error();
   }
 
-  auto img = decodingResult.value;
+  auto img = *decodingResult;
+  if (!img) {
+    // Can happen if missing tiled image is decoded in non-strict mode.
+    return Error(heif_error_Decoder_plugin_error, heif_suberror_Unspecified);
+  }
 
   std::shared_ptr<HeifFile> file = m_heif_context->get_heif_file();
 
@@ -792,8 +725,8 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
 
   if (options.ignore_transformations == false) {
     Result<std::vector<std::shared_ptr<Box>>> propertiesResult = get_properties();
-    if (propertiesResult.error) {
-      return propertiesResult.error;
+    if (!propertiesResult) {
+      return propertiesResult.error();
     }
 
     const std::vector<std::shared_ptr<Box>>& properties = *propertiesResult;
@@ -801,21 +734,21 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
     for (const auto& property : properties) {
       if (auto rot = std::dynamic_pointer_cast<Box_irot>(property)) {
         auto rotateResult = img->rotate_ccw(rot->get_rotation_ccw(), m_heif_context->get_security_limits());
-        if (rotateResult.error) {
+        if (!rotateResult) {
           return error;
         }
 
-        img = rotateResult.value;
+        img = *rotateResult;
       }
 
 
       if (auto mirror = std::dynamic_pointer_cast<Box_imir>(property)) {
         auto mirrorResult = img->mirror_inplace(mirror->get_mirror_direction(),
                                                 get_context()->get_security_limits());
-        if (mirrorResult.error) {
+        if (!mirrorResult) {
           return error;
         }
-        img = mirrorResult.value;
+        img = *mirrorResult;
       }
 
 
@@ -846,11 +779,11 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
           }
 
           auto cropResult = img->crop(left, right, top, bottom, m_heif_context->get_security_limits());
-          if (error) {
-            return error;
+          if (!cropResult) {
+            return cropResult.error();
           }
 
-          img = cropResult.value;
+          img = *cropResult;
         }
       }
     }
@@ -866,12 +799,16 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
 
   std::shared_ptr<ImageItem> alpha_image = get_alpha_channel();
   if (alpha_image) {
-    auto alphaDecodingResult = alpha_image->decode_image(options, decode_tile_only, tile_x0, tile_y0);
-    if (alphaDecodingResult.error) {
-      return alphaDecodingResult.error;
+    if (alpha_image->get_item_error()) {
+      return alpha_image->get_item_error();
     }
 
-    std::shared_ptr<HeifPixelImage> alpha = alphaDecodingResult.value;
+    auto alphaDecodingResult = alpha_image->decode_image(options, decode_tile_only, tile_x0, tile_y0, processed_ids);
+    if (!alphaDecodingResult) {
+      return alphaDecodingResult.error();
+    }
+
+    std::shared_ptr<HeifPixelImage> alpha = *alphaDecodingResult;
 
     // TODO: check that sizes are the same and that we have an Y channel
     // BUT: is there any indication in the standard that the alpha channel should have the same size?
@@ -920,7 +857,7 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
   // Otherwise, use the profile that is stored in the image stream itself and then set the
   // (non-NCLX) profile later.
   auto nclx = get_color_profile_nclx();
-  if (nclx) {
+  if (!nclx.is_undefined()) {
     img->set_color_profile_nclx(nclx);
   }
 
@@ -956,6 +893,20 @@ Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_image(const struct hei
     if (pasp) {
       img->set_pixel_ratio(pasp->hSpacing, pasp->vSpacing);
     }
+
+    // TAI
+
+    auto itai = get_property<Box_itai>();
+    if (itai) {
+      img->set_tai_timestamp(itai->get_tai_timestamp_packet());
+    }
+
+    // GIMI content ID
+
+    auto gimi_content_id = get_property<Box_gimi_content_id>();
+    if (gimi_content_id) {
+      img->set_gimi_sample_content_id(gimi_content_id->get_content_id());
+    }
   }
 
   return img;
@@ -976,44 +927,33 @@ Result<std::vector<uint8_t>> ImageItem::read_bitstream_configuration_data_overri
 }
 #endif
 
-Result<std::vector<uint8_t>> ImageItem::get_compressed_image_data() const
+Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_compressed_image(const heif_decoding_options& options,
+                                                                           bool decode_tile_only, uint32_t tile_x0, uint32_t tile_y0,
+                                                                           std::set<heif_item_id> processed_ids) const
 {
-  // TODO: Remove this later when decoding is done through Decoder.
-
-  // --- get the compressed image data
-
-  // data from configuration blocks
-
-  Result<std::vector<uint8_t>> confData = read_bitstream_configuration_data();
-  if (confData.error) {
-    return confData.error;
+  if (processed_ids.contains(m_id)) {
+    return Error{heif_error_Invalid_input,
+                 heif_suberror_Unspecified,
+                 "'iref' has cyclic references"};
   }
 
-  std::vector<uint8_t> data = confData.value;
-
-  // image data, usually from 'mdat'
-
-  Error error = m_heif_context->get_heif_file()->append_data_from_iloc(m_id, data);
-  if (error) {
-    return error;
-  }
-
-  return data;
-}
+  processed_ids.insert(m_id);
 
 
-Result<std::shared_ptr<HeifPixelImage>> ImageItem::decode_compressed_image(const struct heif_decoding_options& options,
-                                                                           bool decode_tile_only, uint32_t tile_x0, uint32_t tile_y0) const
-{
   DataExtent extent;
   extent.set_from_image_item(get_file(), get_id());
 
-  auto decoder = get_decoder();
-  assert(decoder);
+  auto decoderResult = get_decoder();
+  if (!decoderResult) {
+    return decoderResult.error();
+  }
+
+  auto decoder = *decoderResult;
 
   decoder->set_data_extent(std::move(extent));
 
-  return decoder->decode_single_frame_from_compressed_data(options);
+  return decoder->decode_single_frame_from_compressed_data(options,
+                                                           get_context()->get_security_limits());
 }
 
 
@@ -1058,11 +998,29 @@ Result<std::vector<std::shared_ptr<Box>>> ImageItem::get_properties() const
 }
 
 
+bool ImageItem::has_essential_property_other_than(const std::set<uint32_t>& props) const
+{
+  Result<std::vector<std::shared_ptr<Box>>> propertiesResult = get_properties();
+  if (!propertiesResult) {
+    return false;
+  }
+
+  for (const auto& property : *propertiesResult) {
+    if (is_property_essential(property) &&
+        props.find(property->get_short_type()) == props.end()) {
+      return true;
+    }
+  }
+
+  return false;
+}
+
+
 Error ImageItem::process_image_transformations_on_tiling(heif_image_tiling& tiling) const
 {
   Result<std::vector<std::shared_ptr<Box>>> propertiesResult = get_properties();
-  if (propertiesResult.error) {
-    return propertiesResult.error;
+  if (!propertiesResult) {
+    return propertiesResult.error();
   }
 
   const std::vector<std::shared_ptr<Box>>& properties = *propertiesResult;
